# -*- coding: utf-8 -*-

VOICE_DATA = [
    {
        "Id": "Joanna",
        "LanguageCode": "en-US",
        "LanguageName": "US English",
        "Gender": "Female",
        "Name": "Joanna",
    },
    {
        "Id": "Mizuki",
        "LanguageCode": "ja-JP",
        "LanguageName": "Japanese",
        "Gender": "Female",
        "Name": "Mizuki",
    },
    {
        "Id": "Filiz",
        "LanguageCode": "tr-TR",
        "LanguageName": "Turkish",
        "Gender": "Female",
        "Name": "Filiz",
    },
    {
        "Id": "Astrid",
        "LanguageCode": "sv-SE",
        "LanguageName": "Swedish",
        "Gender": "Female",
        "Name": "Astrid",
    },
    {
        "Id": "Tatyana",
        "LanguageCode": "ru-RU",
        "LanguageName": "Russian",
        "Gender": "Female",
        "Name": "Tatyana",
    },
    {
        "Id": "Maxim",
        "LanguageCode": "ru-RU",
        "LanguageName": "Russian",
        "Gender": "Male",
        "Name": "Maxim",
    },
    {
        "Id": "Carmen",
        "LanguageCode": "ro-RO",
        "LanguageName": "Romanian",
        "Gender": "Female",
        "Name": "Carmen",
    },
    {
        "Id": "Ines",
        "LanguageCode": "pt-PT",
        "LanguageName": "Portuguese",
        "Gender": "Female",
        "Name": "Inês",
    },
    {
        "Id": "Cristiano",
        "LanguageCode": "pt-PT",
        "LanguageName": "Portuguese",
        "Gender": "Male",
        "Name": "Cristiano",
    },
    {
        "Id": "Vitoria",
        "LanguageCode": "pt-BR",
        "LanguageName": "Brazilian Portuguese",
        "Gender": "Female",
        "Name": "Vitória",
    },
    {
        "Id": "Ricardo",
        "LanguageCode": "pt-BR",
        "LanguageName": "Brazilian Portuguese",
        "Gender": "Male",
        "Name": "Ricardo",
    },
    {
        "Id": "Maja",
        "LanguageCode": "pl-PL",
        "LanguageName": "Polish",
        "Gender": "Female",
        "Name": "Maja",
    },
    {
        "Id": "Jan",
        "LanguageCode": "pl-PL",
        "LanguageName": "Polish",
        "Gender": "Male",
        "Name": "Jan",
    },
    {
        "Id": "Ewa",
        "LanguageCode": "pl-PL",
        "LanguageName": "Polish",
        "Gender": "Female",
        "Name": "Ewa",
    },
    {
        "Id": "Ruben",
        "LanguageCode": "nl-NL",
        "LanguageName": "Dutch",
        "Gender": "Male",
        "Name": "Ruben",
    },
    {
        "Id": "Lotte",
        "LanguageCode": "nl-NL",
        "LanguageName": "Dutch",
        "Gender": "Female",
        "Name": "Lotte",
    },
    {
        "Id": "Liv",
        "LanguageCode": "nb-NO",
        "LanguageName": "Norwegian",
        "Gender": "Female",
        "Name": "Liv",
    },
    {
        "Id": "Giorgio",
        "LanguageCode": "it-IT",
        "LanguageName": "Italian",
        "Gender": "Male",
        "Name": "Giorgio",
    },
    {
        "Id": "Carla",
        "LanguageCode": "it-IT",
        "LanguageName": "Italian",
        "Gender": "Female",
        "Name": "Carla",
    },
    {
        "Id": "Karl",
        "LanguageCode": "is-IS",
        "LanguageName": "Icelandic",
        "Gender": "Male",
        "Name": "Karl",
    },
    {
        "Id": "Dora",
        "LanguageCode": "is-IS",
        "LanguageName": "Icelandic",
        "Gender": "Female",
        "Name": "Dóra",
    },
    {
        "Id": "Mathieu",
        "LanguageCode": "fr-FR",
        "LanguageName": "French",
        "Gender": "Male",
        "Name": "Mathieu",
    },
    {
        "Id": "Celine",
        "LanguageCode": "fr-FR",
        "LanguageName": "French",
        "Gender": "Female",
        "Name": "Céline",
    },
    {
        "Id": "Chantal",
        "LanguageCode": "fr-CA",
        "LanguageName": "Canadian French",
        "Gender": "Female",
        "Name": "Chantal",
    },
    {
        "Id": "Penelope",
        "LanguageCode": "es-US",
        "LanguageName": "US Spanish",
        "Gender": "Female",
        "Name": "Penélope",
    },
    {
        "Id": "Miguel",
        "LanguageCode": "es-US",
        "LanguageName": "US Spanish",
        "Gender": "Male",
        "Name": "Miguel",
    },
    {
        "Id": "Enrique",
        "LanguageCode": "es-ES",
        "LanguageName": "Castilian Spanish",
        "Gender": "Male",
        "Name": "Enrique",
    },
    {
        "Id": "Conchita",
        "LanguageCode": "es-ES",
        "LanguageName": "Castilian Spanish",
        "Gender": "Female",
        "Name": "Conchita",
    },
    {
        "Id": "Geraint",
        "LanguageCode": "en-GB-WLS",
        "LanguageName": "Welsh English",
        "Gender": "Male",
        "Name": "Geraint",
    },
    {
        "Id": "Salli",
        "LanguageCode": "en-US",
        "LanguageName": "US English",
        "Gender": "Female",
        "Name": "Salli",
    },
    {
        "Id": "Kimberly",
        "LanguageCode": "en-US",
        "LanguageName": "US English",
        "Gender": "Female",
        "Name": "Kimberly",
    },
    {
        "Id": "Kendra",
        "LanguageCode": "en-US",
        "LanguageName": "US English",
        "Gender": "Female",
        "Name": "Kendra",
    },
    {
        "Id": "Justin",
        "LanguageCode": "en-US",
        "LanguageName": "US English",
        "Gender": "Male",
        "Name": "Justin",
    },
    {
        "Id": "Joey",
        "LanguageCode": "en-US",
        "LanguageName": "US English",
        "Gender": "Male",
        "Name": "Joey",
    },
    {
        "Id": "Ivy",
        "LanguageCode": "en-US",
        "LanguageName": "US English",
        "Gender": "Female",
        "Name": "Ivy",
    },
    {
        "Id": "Raveena",
        "LanguageCode": "en-IN",
        "LanguageName": "Indian English",
        "Gender": "Female",
        "Name": "Raveena",
    },
    {
        "Id": "Emma",
        "LanguageCode": "en-GB",
        "LanguageName": "British English",
        "Gender": "Female",
        "Name": "Emma",
    },
    {
        "Id": "Brian",
        "LanguageCode": "en-GB",
        "LanguageName": "British English",
        "Gender": "Male",
        "Name": "Brian",
    },
    {
        "Id": "Amy",
        "LanguageCode": "en-GB",
        "LanguageName": "British English",
        "Gender": "Female",
        "Name": "Amy",
    },
    {
        "Id": "Russell",
        "LanguageCode": "en-AU",
        "LanguageName": "Australian English",
        "Gender": "Male",
        "Name": "Russell",
    },
    {
        "Id": "Nicole",
        "LanguageCode": "en-AU",
        "LanguageName": "Australian English",
        "Gender": "Female",
        "Name": "Nicole",
    },
    {
        "Id": "Vicki",
        "LanguageCode": "de-DE",
        "LanguageName": "German",
        "Gender": "Female",
        "Name": "Vicki",
    },
    {
        "Id": "Marlene",
        "LanguageCode": "de-DE",
        "LanguageName": "German",
        "Gender": "Female",
        "Name": "Marlene",
    },
    {
        "Id": "Hans",
        "LanguageCode": "de-DE",
        "LanguageName": "German",
        "Gender": "Male",
        "Name": "Hans",
    },
    {
        "Id": "Naja",
        "LanguageCode": "da-DK",
        "LanguageName": "Danish",
        "Gender": "Female",
        "Name": "Naja",
    },
    {
        "Id": "Mads",
        "LanguageCode": "da-DK",
        "LanguageName": "Danish",
        "Gender": "Male",
        "Name": "Mads",
    },
    {
        "Id": "Gwyneth",
        "LanguageCode": "cy-GB",
        "LanguageName": "Welsh",
        "Gender": "Female",
        "Name": "Gwyneth",
    },
    {
        "Id": "Jacek",
        "LanguageCode": "pl-PL",
        "LanguageName": "Polish",
        "Gender": "Male",
        "Name": "Jacek",
    },
]

# {...} is also shorthand set syntax
LANGUAGE_CODES = {
    "cy-GB",
    "da-DK",
    "de-DE",
    "en-AU",
    "en-GB",
    "en-GB-WLS",
    "en-IN",
    "en-US",
    "es-ES",
    "es-US",
    "fr-CA",
    "fr-FR",
    "is-IS",
    "it-IT",
    "ja-JP",
    "nb-NO",
    "nl-NL",
    "pl-PL",
    "pt-BR",
    "pt-PT",
    "ro-RO",
    "ru-RU",
    "sv-SE",
    "tr-TR",
}

VOICE_IDS = {
    "Geraint",
    "Gwyneth",
    "Mads",
    "Naja",
    "Hans",
    "Marlene",
    "Nicole",
    "Russell",
    "Amy",
    "Brian",
    "Emma",
    "Raveena",
    "Ivy",
    "Joanna",
    "Joey",
    "Justin",
    "Kendra",
    "Kimberly",
    "Salli",
    "Conchita",
    "Enrique",
    "Miguel",
    "Penelope",
    "Chantal",
    "Celine",
    "Mathieu",
    "Dora",
    "Karl",
    "Carla",
    "Giorgio",
    "Mizuki",
    "Liv",
    "Lotte",
    "Ruben",
    "Ewa",
    "Jacek",
    "Jan",
    "Maja",
    "Ricardo",
    "Vitoria",
    "Cristiano",
    "Ines",
    "Carmen",
    "Maxim",
    "Tatyana",
    "Astrid",
    "Filiz",
}
